#include <stdlib.h>
#include <stdio.h>

#include "Permute.h"

int CreateOrderTree(struct ordertree *Tree, const long NrNets, const long Lo, const long Hi, const long *Net) {
    long t, NrEmpty;
    
    if (Tree == NULL) {
        fprintf(stderr, "CreateOrderTree(): Null argument!\n");
        return FALSE;
    }
    
    Tree->NetAssign = (int *)malloc(NrNets*sizeof(int));
    Tree->NetToRange = (long *)malloc(NrNets*sizeof(long));
    Tree->Pi = (long *)malloc(NrNets*sizeof(long));
    Tree->Ranges = (struct ordertreerange *)malloc(NrNets*sizeof(struct ordertreerange));
    Tree->TagList = (struct ordertreerange **)malloc(NrNets*sizeof(struct ordertreerange *));
    
    if (Tree->NetAssign == NULL || Tree->NetToRange == NULL || Tree->Pi == NULL || Tree->Ranges == NULL || Tree->TagList == NULL) {
        fprintf(stderr, "CreateOrderTree(): Not enough memory!\n");
        return FALSE;
    }
    
    Tree->NrRanges = 1;
    
    Tree->Ranges[0].PiLo = 0;
    Tree->Ranges[0].PiHi = NrNets;
    Tree->Ranges[0].Meta = 0;
    
    Tree->NrNets = NrNets;
    Tree->NrTags = 0;
    
    for (t = 0; t < NrNets; ++t) {
        Tree->NetToRange[t] = 0;
        Tree->Pi[t] = t;
    }
    
    /* Find empty nets. */
    for (t = 0; t < NrNets; ++t) Tree->NetAssign[t] = 0;
    for (t = Lo; t < Hi; ++t) Tree->NetAssign[Net[t]] = 1;
    
    /* Permute them to the back. */
    NrEmpty = 0;
    
    for (t = 0; t < NrNets; ++t) {
        if (Tree->NetAssign[t] == 0) Tree->Pi[NrNets - (++NrEmpty)] = t;
    }
    
    Tree->Ranges[0].PiHi = NrNets - NrEmpty;
    
    /* Permute nonempty nets to the front. */
    NrEmpty = 0;
    
    for (t = 0; t < NrNets; ++t) {
        if (Tree->NetAssign[t] == 1) Tree->Pi[NrEmpty++] = t;
    }
    
    return TRUE;
}

int DestroyOrderTree(struct ordertree *Tree) {
    if (Tree == NULL) {
        fprintf(stderr, "DestroyOrderTree(): Null argument!\n");
        return FALSE;
    }
    
    if (Tree->NetAssign != NULL) free(Tree->NetAssign);
    if (Tree->NetToRange != NULL) free(Tree->NetToRange);
    if (Tree->Pi != NULL) free(Tree->Pi);
    if (Tree->Ranges != NULL) free(Tree->Ranges);
    if (Tree->TagList != NULL) free(Tree->TagList);
    
    memset(Tree, 0, sizeof(struct ordertree));
    
    return TRUE;
}

int remembrance_add( struct remembrance * const m, const long x ) {
    if( m == NULL ) {
        fprintf( stderr, "remembrance_get(): Null pointer supplied!\n" );
        return FALSE;
    }

    /* Adaptively grow vector. */
    if( m->size == m->alloc ) {
        m->alloc *= 2;
        m->vector = realloc( m->vector, m->alloc*sizeof( long ) );
        if( m->vector == NULL ) {
            fprintf(stderr, "remembrance_add(): Error during realloc of vector\n");
            return FALSE;
        }
    }

    /* Add element and increment size field. */
    m->vector[ m->size++ ] = x;
    return TRUE;
}

int OrderTreeSplit(struct ordertree *Tree, const long Lo, const long Mid, const long Hi, const long *Net, const struct opts *pOptions, struct remembrance * const m ) {
    long t;
    
    if (Tree == NULL || Net == NULL || pOptions == NULL) {
        fprintf(stderr, "OrderTreeSplit(): Null arguments!\n");
        return FALSE;
    }
    
    /* printf("Start with %ld/%ld ranges, split %ld - %ld - %ld.\n", Tree->NrRanges, Tree->NrNets, Lo, Mid, Hi); */
    
    /* Clear all net assignments and mark all ranges as untagged. */
    Tree->NrTags = 0;
    
    for (t = Lo; t < Hi; ++t) {
        const long n = Net[t];
        struct ordertreerange *r = &Tree->Ranges[Tree->NetToRange[n]];
        
        Tree->NetAssign[n] = 0;
        
        r->LeftCount = r->CutCount = r->RightCount = 0;
        r->LeftChild = r->CutChild = r->RightChild = -1;
        r->WasTagged = FALSE;
    }
    
    /* Assign all left nets and permute them to the beginning. */
    for (t = Lo; t < Mid; ++t) {
        const long n = Net[t];
        struct ordertreerange *r = &Tree->Ranges[Tree->NetToRange[n]];
        
        if (Tree->NetAssign[n] == 0) {
            Tree->NetAssign[n] = 1;
            r->LeftCount++;
            
            if (!r->WasTagged) {
                r->WasTagged = TRUE;
                Tree->TagList[Tree->NrTags++] = r;
            }
        }
    }
    
    /* Assign all right nets and permute them to the end. */
    for (t = Mid; t < Hi; ++t) {
        const long n = Net[t];
        struct ordertreerange *r = &Tree->Ranges[Tree->NetToRange[n]];
        
        if (Tree->NetAssign[n] == 0) {
            Tree->NetAssign[n] = 2;
            r->RightCount++;
            
            if (!r->WasTagged) {
                r->WasTagged = TRUE;
                Tree->TagList[Tree->NrTags++] = r;
            }
        }
        else if (Tree->NetAssign[n] == 1) {
            /* This net is cut: it should be permuted to the middle. If this is the case, then this range has already been tagged! */
            Tree->NetAssign[n] = 3;
            r->LeftCount--;
            r->CutCount++;
        }
    }
    
    /* Update net-to-range mapping. */
    /* Check ranges and set starting offsets. */
    for (t = 0; t < Tree->NrTags; ++t) {
        struct ordertreerange *r = Tree->TagList[t];
        
        if (!r->WasTagged) {
            fprintf(stderr, "OrderTreeSplit(): Sanity check I failed!\n");
            return FALSE;
        }
        
        if (r->PiHi != r->PiLo + r->LeftCount + r->CutCount + r->RightCount) {
            fprintf(stderr, "OrderTreeSplit(): Sanity check II failed!\n");
            return FALSE;
        }
       
        switch (pOptions->OrderPermutation) {
        case OrderPrefix:
            r->CutStart = r->PiLo;
            r->LeftStart = r->PiLo + r->CutCount;
            r->RightStart = r->PiLo + r->CutCount + r->LeftCount;
            break;
        case OrderPostfix:
            r->LeftStart = r->PiLo;
            r->RightStart = r->PiLo + r->LeftCount;
            r->CutStart = r->PiLo + r->LeftCount + r->RightCount;
            break;
        case OrderInfix:
        default:
            r->LeftStart = r->PiLo;
            r->CutStart = r->PiLo + r->LeftCount;
            r->RightStart = r->PiLo + r->LeftCount + r->CutCount;
            break;
        }
    }
    
    /* Now we determine which ranges need to be split and where their children will spawn. */
    /* Because the original range will be overwritten, the range child index may be equal to the range index. */
    /* Ranges will be overwritten with priority Left, Cut, Right (so if the left range is nonempty. this will overwrite the current range, otherwise the cut range will overwrite the current range and if both the left and cut ranges are empty, the right range will overwrite the current range. */
    for (t = Lo; t < Hi; ++t) {
        const long n = Net[t], i = Tree->NetToRange[n];
        struct ordertreerange *r = &Tree->Ranges[i];
        
        /* Only recalculate the appropriate range *once* per net. */
        if (Tree->NetAssign[n] == 4) continue;
        
        switch (Tree->NetAssign[n]) {
        case 1:
            if (r->LeftChild >= 0) Tree->NetToRange[n] = r->LeftChild;
            else r->LeftChild = i;
            
            Tree->Pi[r->LeftStart++] = n;
            break;
        case 2:
            if (r->RightChild >= 0) Tree->NetToRange[n] = r->RightChild;
            else if (r->LeftCount == 0 && r->CutCount == 0) r->RightChild = i;
            else {
                Tree->NetToRange[n] = r->RightChild = Tree->NrRanges++;
                Tree->Ranges[r->RightChild].Meta=0;
            }
            
            Tree->Pi[r->RightStart++] = n;
            break;
        case 3:
            if (r->CutChild >= 0) Tree->NetToRange[n] = r->CutChild;
            else if (r->LeftCount == 0) r->CutChild = i;
            else {
                Tree->NetToRange[n] = r->CutChild = Tree->NrRanges++;
                Tree->Ranges[r->CutChild].Meta=0;
            }
            
            Tree->Pi[r->CutStart++] = n;
            break;
        default:
            fprintf(stderr, "OrderTreeSplit(): Unassigned net!\n");
            return FALSE;
            break;
        }
        
        Tree->NetAssign[n] = 4;
    }
    
    /* Now actually split the ranges. */
    for (t = 0; t < Tree->NrTags; ++t) {
        const struct ordertreerange r = *Tree->TagList[t];
        struct ordertreerange *s;
        
        /* printf("Children %ld (count %ld) - %ld (count %ld) - %ld (count %ld), permutation range %ld - %ld, tagged %d.\n", r.LeftChild, r.LeftCount, r.CutChild, r.CutCount, r.RightChild, r.RightCount, r.PiLo, r.PiHi, r.WasTagged); */
        
        if (r.LeftChild >= 0) {
            s = &Tree->Ranges[r.LeftChild];
            s->Meta = r.Meta;
            
            switch (pOptions->OrderPermutation) {
            case OrderPrefix:
                s->PiLo = r.PiLo + r.CutCount;
                s->PiHi = r.PiLo + r.CutCount + r.LeftCount;
                /* Remembrance: in prefix, both s->PiLo and s->PiHi will be remembered later (if applicable!) */
                break;
            case OrderPostfix:
            case OrderInfix:
            default:
                s->PiLo = r.PiLo;
                s->PiHi = r.PiLo + r.LeftCount;
                /* Remembrance, both in postfix and infix, s->PiHi will be remembered later (if applicable!) */
                break;
            }
        }
        
        if (r.CutChild >= 0) {
            s = &Tree->Ranges[r.CutChild];
            
            s->Meta = ( r.Meta==0 ? 1 : 2 );

            switch (pOptions->OrderPermutation) {
            case OrderPrefix:
                s->PiLo = r.PiLo;
                s->PiHi = r.PiLo + r.CutCount;
                if( s->Meta < 2 )
                    if( remembrance_add( m, s->PiHi )==FALSE ) {
                        fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                        return FALSE;
                    }
                break;
            case OrderPostfix:
                s->PiLo = r.PiLo + r.LeftCount + r.RightCount;
                s->PiHi = r.PiLo + r.LeftCount + r.RightCount + r.CutCount;
                if( s->Meta < 2 ) {
                    if( remembrance_add( m, s->PiLo )==FALSE ) {
                        fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                        return FALSE;
                    }
                }
                break;
            case OrderInfix:
            default:
                s->PiLo = r.PiLo + r.LeftCount;
                s->PiHi = r.PiLo + r.LeftCount + r.CutCount;
                if( s->Meta < 2 ) {
                        if( remembrance_add( m, s->PiLo )==FALSE ) {
                            fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                            return FALSE;
                        }
                        if( remembrance_add( m, s->PiHi )==FALSE ) {
                            fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                            return FALSE;
                        }
                }
                break;
            }
        }

        if( r.CutCount == 0 && r.Meta == 0 ) {
            /* Remembrance: empty cut! */
            switch( pOptions->OrderPermutation) {
            case OrderPostfix:
                remembrance_add( m, r.PiHi );
                break;
            case OrderInfix:
                remembrance_add( m, r.PiLo + r.LeftCount );
                remembrance_add( m, r.PiLo + r.LeftCount );
                break;
            case OrderPrefix:
                remembrance_add( m, r.PiLo );
                break;
            default:
                fprintf( stderr, "Undefined permutation order\n" );
                return FALSE;
            }
        }
        
        if (r.RightChild >= 0) {
            s = &Tree->Ranges[r.RightChild];
            s->Meta = r.Meta;

            switch (pOptions->OrderPermutation) {
            case OrderPostfix:
                s->PiLo = r.PiLo + r.LeftCount;
                s->PiHi = r.PiLo + r.LeftCount + r.RightCount;
                if( s->Meta < 1 ) { /* Be aware: 2 would be faulty */
                    if( remembrance_add( m, s->PiLo )==FALSE ) {
                        fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                        return FALSE;
                    }
                }
                /* Remembrance: in postfix, s->PiHi is already remembered. */
                break;
            case OrderPrefix:
                s->PiLo = r.PiLo + r.LeftCount + r.CutCount;
                s->PiHi = r.PiLo + r.LeftCount + r.CutCount + r.RightCount;
                if( s->Meta < 1 ) { /* Be aware: 2 would be faulty */
                    if( remembrance_add( m, s->PiLo )==FALSE ) {
                        fprintf(stderr, "OrderTreeSplit(): Remembrance failed!\n");
                        return FALSE;
                    }
                }
                break;
            case OrderInfix:
            default:
                s->PiLo = r.PiLo + r.LeftCount + r.CutCount;
                s->PiHi = r.PiLo + r.LeftCount + r.CutCount + r.RightCount;
                /* Remembrance: in infix, s->PiLo is already remembered. */
                break;
            }
        }
    }
    
    return TRUE;
}

