/*
MatlabMondriaan.c

Created by Bas Fagginger Auer, based on the work done by Ken Stanley.

This is a Matlab wrapper for the Mondriaan sparse matrix library which takes as input a (sparse) matrix from Matlab and outputs the same matrix, but with all non-zero entries replaced by the corresponding processor number to which the non-zero entry has been assigned by the Mondriaan algorithm.

Usage: 	[B, s, u, v] = MatlabMondriaan(A, p, epsilon, maxiter, permute)
	
	Distributes the non-zero entries of A among p processors with imbalance epsilon using the Mondriaan algorithm.
	B(i, j) = # of processor to which the entry A(i, j) has been assigned by Mondriaan, between 1 and p.
	Similarly u(i) and v(i) = # of processor to which component i of u, resp. v is assigned.
	s contains the statistics of the partitioning: distribution time, imbalance, maximum communication, and communication volume.
	The variable maxiter contains the maximum number of iterations the algorithm should perform before exiting, useful to produce animations.
	The variable permute contains an integer which sets the desired permutation: 0 for no permutation, 1 for reverse BBD, 2 for SBD, and 3 for BBD.
*/
#include <stdlib.h>
#include <time.h>
#include <Mondriaan.h>

#ifndef USE_MATLAB
#error To be able to use the Matlab interface of Mondriaan, please make the appropriate changes in ../mondriaan.mk.
#endif

/* Matlab include files. */
#include "mex.h"
#include "matrix.h"

/* For compatibility with older versions of Matlab, we need to define mwSize and mwIndex ourselves. */
#ifndef mwSize
#define mwSize size_t
#endif
#ifndef mwIndex
#define mwIndex size_t
#endif

/*
Let us take a look at the Matlab sparse matrix format (see mxSetJc in the Matlab help).
Denote the row indices by RI, column indices by CI, and data by D.
Then the RI array contains the indices of the rows that contain non-zero elements offset by 1, so if RI[0] = 0, then the first row of the matrix contains a non-zero element.
The D array contains the values of the non-zero entries.
Finally the CI array is the most complicated: CI[i] is the index in RI and D of the first non-zero entry in the i-th column.

So if we would have the 16x14 matrix A with non-zero entries A(2, 2) = 9, A(4, 2) = 7, A(13, 5) = 2, then
RI = 2 4 13
D = 9 7 2
CI = 0 0 2 2 2 3 (no elements in first column, two elements in second column, no elements in third and fourth column, one if fifth column).
*/

/* This structure hold the partitioning statistics. */
struct MondriaanStats {
	/* Maximum and minimum number of nonzeros assigned to a single processor. */
	long MaxNz, MinNz;
	double Epsilon;
	long MaxComU, MaxComV;
	long ComVolU, ComVolV;
	double Duration;
};

/* This structure will be used to translate a Mondriaan sparse matrix back to the Matlab format. */
struct MondriaanTriplet {
	long Row, Column;
	double ReValue, ImValue;
};

int MaxIterations = -1;

/* This callback ensures that we stop after MaxIterations iterations. */
int Callback(int Iteration, int SplitPart, const struct sparsematrix *A) {
	if (MaxIterations <= 0) return true;
	
	return (Iteration < MaxIterations);
}

/* This function runs the Mondriaan algorithm on a given sparse matrix A. */
int DoMondriaan(struct sparsematrix *A, int *u, int *v, struct MondriaanStats *s, int NumProcessors, double Imbalance, int Permutation) {
	clock_t Clock;
	int Symmetric;
	struct opts Options;
	int i;
	long l;
	
	if (A == NULL || s == NULL || u == NULL || v == NULL) return false;
	
	/* Read Mondriaan options. */
	SetDefaultOptions(&Options);
	SetOptionsFromFile(&Options, "Mondriaan.defaults");
	
	/* Set desired permutation ordering. */
	if (Permutation == 1) Options.OrderPermutation = OrderPrefix;
	else if (Permutation == 2) Options.OrderPermutation = OrderInfix;
	else if (Permutation == 3) Options.OrderPermutation = OrderPostfix;
	
	Options.matrix = NULL;
	Options.P = NumProcessors;
	Options.eps = Imbalance;
	
	if (!ApplyOptions(&Options)) return false;
	
	/* Start the timer. */
	Clock = clock();
	
	/* Remove double zeroes. */
	if (!SparseMatrixRemoveDuplicates(A)) return false;
	
	/* We do not have to take weighted matrices into account, because we are working in Matlab. */
	
	/* Check symmetry. */
	if (A->m == A->n && (A->MMTypeCode[3] == 'S' || A->MMTypeCode[3] == 'K' || A->MMTypeCode[3] == 'H')) Symmetric = true;
	else Symmetric = false;
	
	/* Convert to symmetric form if so desired. */
	if (Symmetric)
	{
		if (Options.SymmetricMatrix_UseSingleEntry == SingleEntNo) SparseMatrixSymmetric2Full(A);
		else if (Options.SymmetricMatrix_SingleEntryType == ETypeRandom) SparseMatrixSymmetricLower2Random(A);
	}
	
	/* Add dummies if requested. */
	if (A->m == A->n && Options.SquareMatrix_DistributeVectorsEqual == EqVecYes && Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes) AddDummiesToSparseMatrix(A);
	
	/* Set up parameters of A. */
	A->NrProcs = NumProcessors;
	A->Pstart = (long *)malloc((A->NrProcs + 1)*sizeof(long));
	
	if (A->Pstart == NULL) return false;
	
	/* Initialise processor array. */	
	A->Pstart[0] = 0;
	
	for (i = 1; i <= A->NrProcs; i++) A->Pstart[i] = A->NrNzElts;
	
	/* Distribute the processors among the matrix entries. */
	if (!DistributeMatrixMondriaan(A, NumProcessors, Imbalance, &Options, MaxIterations <= 0 ? NULL : Callback)) return false;
	
	/* Remove dummies. */
	if (A->m == A->n && Options.SquareMatrix_DistributeVectorsEqual == EqVecYes && Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes) RemoveDummiesFromSparseMatrix(A);
	
	/* Restore symmetry. */
	if (Symmetric && Options.SymmetricMatrix_UseSingleEntry == SingleEntYes && Options.SymmetricMatrix_SingleEntryType == ETypeRandom) SparseMatrixSymmetricRandom2Lower(A);
	
	/* Substitute processor numbers for non-zero entries and delete imaginary part. */
	if (!MMInsertProcessorIndices(A)) return false;
	
	/* Convert to full form for vector partitioning. */
	if (Symmetric && Options.SymmetricMatrix_UseSingleEntry == SingleEntYes) SparseMatrixSymmetric2Full(A);
	
	/* Distribute vectors. */
	if (A->m == A->n && Options.SquareMatrix_DistributeVectorsEqual == EqVecYes) {
		/* Distribute vectors equally. */
		if (Symmetric && Options.SymmetricMatrix_UseSingleEntry == SingleEntYes) {
			if ((s->MaxComV = DistributeVec(A, v, ROW, &Options)) < 0) return false;
			
			for (i = 0; i < A->m; i++) u[i] = v[i];
			
			s->MaxComU = s->MaxComV;
		}
		else {
			if ((s->MaxComU = DistributeVecOrigEq(A, u, v, &Options)) < 0) return false;
			
			s->MaxComV = 0;
		}
	}
	else {
		/* Distribute vectors independently. */
		if ((s->MaxComV = DistributeVec(A, v, ROW, &Options)) < 0 || (s->MaxComU = DistributeVec(A, u, COL, &Options)) < 0) return false;
	}
	
	/* Calculate duration of partitioning. */
	s->Duration = (double)(clock() - Clock)/(double)CLOCKS_PER_SEC;
	
	/* Determine further statistics. */
	s->MinNz = s->MaxNz = A->Pstart[1] - A->Pstart[0];
	
	for (i = 1; i < A->NrProcs; i++) {
		l = A->Pstart[i + 1] - A->Pstart[i];
		
		if (l > s->MaxNz) s->MaxNz = l;
		if (l < s->MinNz) s->MinNz = l;
	}
	
	s->Epsilon = (double)(A->NrProcs*s->MaxNz - A->NrNzElts)/(double)A->NrNzElts;
	
	CalcCom(A, v, ROW, &s->ComVolV, &l, &l, &l, &l);
	CalcCom(A, u, COL, &s->ComVolU, &l, &l, &l, &l);
	
	return true;
}

/* This function tries to extract a double precision value from the given variable. */
double ExtractDouble(const mxArray *Var) {
	size_t NumRows, NumCols;
	
	if (!Var) mexErrMsgTxt("Null variable pointer!");
	
	NumRows = mxGetM(Var);
	NumCols = mxGetN(Var);
	
	if (mxIsComplex(Var) || NumRows != 1 || NumCols != 1) mexErrMsgTxt("Provided variable is not a real number!");
	
	return mxGetScalar(Var);
}

/* This function converts a Matlab sparse matrix to a Mondriaan sparse matrix. */
struct sparsematrix *ConvertMatlabToMondriaan(const mxArray *A) {
	struct sparsematrix *B;
	mwIndex *AIr, *AJc;
	double *APr, *APi;
	long i, j, k;
	
	/* Read input Matlab matrix data. */
	if (A == NULL) mexErrMsgTxt ("Null variable pointer!");
	if (mxGetNumberOfDimensions (A) != 2 || !mxIsSparse (A) || mxGetM (A) <= 1 || mxGetN (A) <= 1) mexErrMsgTxt ("Matrix that we have to retrieve is not a sparse matrix!");
	
	APr = mxGetPr(A);
	APi = (mxIsComplex(A) ? mxGetPi(A) : NULL);
	AIr = mxGetIr(A);
	AJc = mxGetJc(A);
	
	if (!APr || !AIr || !AJc) mexErrMsgTxt ("Matrix data could not be retrieved!");
	
	B = (struct sparsematrix *)malloc(sizeof (struct sparsematrix));
	
	if (B == NULL) mexErrMsgTxt("Could not allocate Mondriaan matrix!");
	
	/* First initialise the sparse matrix to its default parameters. */
	if (!MMSparseMatrixInit(B)) mexErrMsgTxt("Unable to initialise matrix!");
	
	/* Now fill in the parameters from A. */
	B->MMTypeCode[0] = 'M';
	B->MMTypeCode[1] = 'C';
	B->MMTypeCode[2] = (APi != NULL ? 'C' : 'R');
	B->MMTypeCode[3] = 'G';
	
	B->m = mxGetM(A);
	B->n = mxGetN(A);
	B->NrNzElts = AJc[B->n];
	
	/* Allocate arrays to hold the data of all matrix elements. */
	if (!MMSparseMatrixAllocateMemory(B)) mexErrMsgTxt("Unable to allocate matrix!");
	if (B->i == NULL || B->j == NULL || B->ReValue == NULL || (B->ImValue == NULL && APi != NULL)) mexErrMsgTxt ("Unable to allocate Mondriaan matrix arrays!");
	
	/* Fill the arrays with data from A. */
	i = 0;
	
	for (j = 0; j < B->n && AJc[j] < (size_t)B->NrNzElts; j++) {
		for (k = AJc[j + 1] - AJc[j]; k-- > 0; ) {
			B->i[i] = AIr[i];
			B->j[i] = j;
			B->ReValue[i] = APr[i];
			
			if (B->ImValue != NULL) B->ImValue[i] = APi[i];
			
			i++;
		}
	}
	
	return B;
}
/* Compare function necessary to order triplets in a fashion that is compatible with the Matlab sparse matrix format. */
int CompareTriplets(const void *p1, const void *p2) {
	const struct MondriaanTriplet *t1, *t2;
	
	t1 = (struct MondriaanTriplet *)p1;
	t2 = (struct MondriaanTriplet *)p2;
	
	if (t1->Column != t2->Column) return t1->Column - t2->Column;
	
	return t1->Row - t2->Row;
}

/* This function converts Mondriaan statistics to a matlab vector. */
mxArray *ConvertStatsToVector(const struct MondriaanStats *s) {
	mxArray *w;
	double *wp;
	
	if (s == NULL) mexErrMsgTxt("Null stats pointer!");
	
	w = mxCreateDoubleMatrix(4, 1, mxREAL);
	
	if (w == NULL) mexErrMsgTxt("Unable to allocate Matlab vector!");
	
	wp = mxGetPr(w);
	
	wp[0] = s->Duration;
	wp[1] = s->Epsilon;
	wp[2] = (double)(s->MaxComU + s->MaxComV);
	wp[3] = (double)(s->ComVolU + s->ComVolV);
	
	return w;
}

/* This function converts a vector assignment to a vector of appropriate size. */
mxArray *ConvertIndicesToVector(const int *v, const size_t Size) {
	mxArray *w;
	size_t i;
	double *wp;
	
	if (v == NULL) mexErrMsgTxt("Null variable pointer!");
	
	w = mxCreateDoubleMatrix(Size, 1, mxREAL);
	
	if (w == NULL) mexErrMsgTxt("Unable to allocate Matlab vector!");
	
	wp = mxGetPr(w);
	
	for (i = 0; i < Size; i++) wp[i] = (double)(v[i] + 1);
	
	return w;
}

/* This function converts a Mondriaan sparse matrix to a Matlab sparse matrix. */
mxArray *ConvertMondriaanToMatlab(const struct sparsematrix *B) {
	struct MondriaanTriplet *Triplets;
	mxArray *A;
	mwIndex *AIr, *AJc;
	double *APr, *APi;
	int CurrentColumn;
	long i;
	
	if (B == NULL) mexErrMsgTxt("Null variable pointer!");
	
	/* Extract all non-zero entries. */
	Triplets = (struct MondriaanTriplet *)malloc(B->NrNzElts*sizeof (struct MondriaanTriplet));
	
	if (Triplets == NULL) mexErrMsgTxt("Unable to allocate triplets!");
	
	for (i = 0; i < B->NrNzElts; i++) {
		Triplets[i].Row = B->i[i];
		Triplets[i].Column = B->j[i];
		
		if (B->row_perm_inv != NULL) Triplets[i].Row = B->row_perm_inv[B->i[i]];
		if (B->col_perm_inv != NULL) Triplets[i].Column = B->col_perm_inv[B->j[i]];
		
		Triplets[i].ReValue = B->ReValue[i];
		
		if (B->ImValue) Triplets[i].ImValue = B->ImValue[i];
		else Triplets[i].ImValue = 0.0;
	}
	
	/* Sort them in a Matlab friendly way. */
	qsort(Triplets, B->NrNzElts, sizeof (struct MondriaanTriplet), CompareTriplets);
	
	/* Allocate Matlab matrix and obtain pointers to its real data, imaginary data, row index, and column count arrays respectively. */
	A = mxCreateSparse(B->m, B->n, B->NrNzElts, B->ImValue == NULL ? mxREAL : mxCOMPLEX);
	
	if (A == NULL) mexErrMsgTxt("Unable to allocate Matlab matrix!");
	
	APr = mxGetPr (A);
	APi = mxGetPi (A);
	AIr = mxGetIr (A);
	AJc = mxGetJc (A);
	
	/* Write triplets to the Matlab matrix. */
	CurrentColumn = 0;
	*AJc++ = 0;
	
	for (i = 0; i < B->NrNzElts; i++) {
		*AIr++ = Triplets[i].Row;
		*APr++ = Triplets[i].ReValue;
		
		if (APi) *APi++ = Triplets[i].ImValue;
		
		/* Check whether or not we need to advance to the next column. */
		while (CurrentColumn < Triplets[i].Column) {
			*AJc++ = i;
			CurrentColumn++;
		}
	}
	
	*AJc++ = B->NrNzElts;
	
	free(Triplets);
	
	return A;
}

/* This function provides the actual Matlab interface. */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	/* These are variables we need to be able to use Mondriaan. */
	struct sparsematrix *MondriaanMatrix;
	struct MondriaanStats Stats;
	int *MondriaanU, *MondriaanV;
	int NumProcessors, Permutation;
	double Imbalance;
	
	/* Check that the input parameters are valid. */
	if (nrhs != 5) mexErrMsgTxt("We require five input variables: the sparse matrix, the number of processors, the imbalance factor, the maximum number of iterations, and the permutation type!");
	else if (nlhs != 2 && nlhs != 4) mexErrMsgTxt("We require either two or four output variables!");
	else if (!mxIsSparse(prhs[0])) mexErrMsgTxt("The matrix on which we work needs to be sparse!");
	
	/* Extract the number of processors and imbalance and verify their values. */
	NumProcessors = (int)ExtractDouble (prhs[1]);
	Imbalance = ExtractDouble (prhs[2]);
	MaxIterations = (int)ExtractDouble (prhs[3]);
	Permutation = (int)ExtractDouble (prhs[4]);
	
	if (NumProcessors <= 0) mexErrMsgTxt ("The number of processors should be a positive integer!");
	if (Imbalance <= 0.0 || Imbalance >= 1.0) mexErrMsgTxt ("The imbalance should lie between 0 and 1!");
	
	/* Convert matrix to a Mondriaan friendly format. */
	MondriaanMatrix = ConvertMatlabToMondriaan(prhs[0]);
	
	if (MondriaanMatrix == NULL) mexErrMsgTxt("Unable to convert Matlab matrix to Mondriaan!");
	
	/* Run the Mondriaan algorithm. */
	MondriaanU = (int *)malloc(MondriaanMatrix->m*sizeof(int));
	MondriaanV = (int *)malloc(MondriaanMatrix->n*sizeof(int));
	
	if (MondriaanU == NULL || MondriaanV == NULL) mexErrMsgTxt("Not enough memory to allocate vectors!");
	
	if (!DoMondriaan(MondriaanMatrix, MondriaanU, MondriaanV, &Stats, NumProcessors, Imbalance, Permutation)) mexErrMsgTxt("Unable to run the Mondriaan algorithm!");
	
	/* And we convert the Mondriaan matrix back to Matlab. */
	plhs[0] = ConvertMondriaanToMatlab(MondriaanMatrix);
	plhs[1] = ConvertStatsToVector(&Stats);
	
	/* Free the Mondriaan matrix. */
	MMDeleteSparseMatrix(MondriaanMatrix);
	
	if (plhs[0] == NULL || plhs[1] == NULL) mexErrMsgTxt("Unable to convert Mondriaan matrix back to Matlab!");
	
	/* Return vector distributions if desired. */
	if (nlhs == 4) {
		plhs[2] = ConvertIndicesToVector(MondriaanU, MondriaanMatrix->m);
		plhs[3] = ConvertIndicesToVector(MondriaanV, MondriaanMatrix->n);
	
		free(MondriaanU);
		free(MondriaanV);
			
		if (plhs[2] == NULL || plhs[3] == NULL) mexErrMsgTxt("Unable to convert Mondriaan vectors to Matlab!");
	}
}

